"""Version string parser and generator."""

import collections.abc
import enum
import itertools
import logging
import re
import typing as t

import packaging.version
import semver

_LOG = logging.getLogger(__name__)


@enum.unique
class VersionComponent(enum.IntEnum):
    """Enumeration of standard version components."""

    Major = 1 << 1
    Minor = 1 << 2
    Patch = 1 << 3
    Release = Major | Minor | Patch
    PrePatch = 1 << 4
    DevPatch = 1 << 5
    Local = 1 << 6


def _version_tuple_checker(version_tuple, flags):
    return all([(_ is not None if flag else _ is None) for _, flag in zip(version_tuple, flags)])


class Version(collections.abc.Hashable):  # pylint: disable = too-many-public-methods
    """For storing and manipulating version information.

    Definitions of acceptable version formats are provided in readme.
    """

    _re_number = r'(?:0|[123456789][0123456789]*)'
    # _re_sha = r'[0123456789abcdef]+'
    _re_letters = r'(?:[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ]+)'
    _pattern_letters = re.compile(_re_letters)
    _re_alphanumeric = r'(?:[0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ]+)'
    _pattern_alphanumeric = re.compile(_re_alphanumeric)
    _re_sep = r'(?:[\.-])'

    _re_release_parts = \
        r'(?P<major>{n})(?:\.(?P<minor>{n}))?(?:\.(?P<patch>{n}))?'.format(n=_re_number)
    _pattern_release = re.compile(_re_release_parts)

    @classmethod
    def _parse_release_str(cls, release: str) -> tuple:
        match = cls._pattern_release.fullmatch(release)
        assert match is not None
        major_match = match.group('major')
        assert major_match is not None
        major = int(major_match)
        minor_match = match.group('minor')
        if minor_match is not None:
            minor = int(minor_match)
        else:
            minor = None
        patch_match = match.group('patch')
        if patch_match is not None:
            patch = int(patch_match)
        else:
            patch = None
        return major, minor, patch

    _re_pre_separator = rf'(?P<preseparator>{_re_sep})'
    _re_pre_type = rf'(?P<pretype>{_re_letters})'
    _re_pre_patch = rf'(?P<prepatch>{_re_number})'
    _re_pre_release_part = rf'{_re_pre_separator}?{_re_pre_type}?{_re_pre_patch}?'
    _pattern_pre_release_part = re.compile(_re_pre_release_part)
    _re_pre_release_parts = r'(?:{0}{2})|(?:{0}?{1}{2}?)'.format(_re_sep, _re_letters, _re_number)
    _pattern_pre_release = re.compile(_re_pre_release_parts)
    _pattern_pre_release_check = re.compile(rf'(?:{_re_pre_release_parts})+')

    @classmethod
    def _parse_pre_release_str(cls, pre_release: str) -> t.Sequence[
            t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]]:
        parts = cls._pattern_pre_release.findall(pre_release)
        _LOG.debug('parsed pre-release string %s into %s',
                   repr(pre_release), parts)
        tuples = []
        for part in parts:
            match = cls._pattern_pre_release_part.fullmatch(part)
            assert match is not None
            pre_patch_match = match.group('prepatch')
            if pre_patch_match is not None:
                pre_patch = int(pre_patch_match)
            else:
                pre_patch = None
            tuples.append((match.group('preseparator'), match.group('pretype'), pre_patch))
        return tuples

    _re_local_separator = rf'({_re_sep})'
    _re_local_part = rf'({_re_alphanumeric})'
    _re_local_parts = rf'\+{_re_local_part}(?:{_re_local_separator}{_re_local_part})*'
    _pattern_local = re.compile(_re_local_parts)

    @classmethod
    def _parse_local_str(cls, local: str) -> tuple:
        match = cls._pattern_local.fullmatch(local)
        assert match is not None
        return tuple([_ for _ in match.groups() if _ is not None])

    _re_release = r'(?P<release>{n}(?:\.{n})?(?:\.{n})?)'.format(n=_re_number)
    _re_pre_release = r'(?P<prerelease>(?:(?:{0}{2})|(?:{0}?{1}{2}?))+)'.format(
        _re_sep, _re_letters, _re_number)
    _re_local = r'(?P<local>\+{0}([\.-]{0})*)'.format(_re_alphanumeric)
    # _re_named_parts_count = 3 + 3
    _re_version = rf'{_re_release}{_re_pre_release}?{_re_local}?'
    _pattern_version = re.compile(_re_version)

    @classmethod
    def from_str(cls, version_str: str):
        """Create version from string."""
        match = cls._pattern_version.fullmatch(version_str)  # type: t.Optional[t.Match[str]]
        if match is None:
            raise ValueError(f'version string {repr(version_str)} is invalid')
        _LOG.debug('version_query parsed version string %s into %s: %s %s',
                   repr(version_str), type(match), match.groupdict(), match.groups())

        _release = match.group('release')
        _pre_release = match.group('prerelease')
        _local = match.group('local')

        major, minor, patch = cls._parse_release_str(_release)
        pre_release = None if _pre_release is None else cls._parse_pre_release_str(_pre_release)
        local = None if _local is None else cls._parse_local_str(_local)

        return cls(major=major, minor=minor, patch=patch, pre_release=pre_release, local=local)

    @classmethod
    def from_tuple(cls, version_tuple: tuple):
        return cls(*version_tuple)

    @classmethod
    def from_dict(cls, version_dict: dict):
        return cls(**version_dict)

    @classmethod
    def from_py_version(cls, py_version: packaging.version.Version):
        """Create version from a standard Python version object."""
        if not isinstance(py_version, packaging.version.Version):
            _LOG.warning('attempting to parse %s as packaging.version.Version...', type(py_version))
        ver = py_version._version
        major, minor, patch = [ver.release[i] if len(ver.release) > i
                               else None for i in range(3)]
        pre_release: t.Optional[t.Sequence[
            t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]]] = None
        local = None
        pre_ver: t.Optional[t.Tuple[None, int]]
        if len(ver.release) == 4:
            pre_ver = (None, ver.release[3])
        elif len(ver.release) > 4:
            raise NotImplementedError(py_version)
        else:
            pre_ver = None
        pre_ver_present = sum(1 for _ in (ver.post, ver.dev, ver.pre) if _)
        if pre_ver and pre_ver_present:
            raise NotImplementedError(py_version)
        if pre_ver_present > 1:
            raise NotImplementedError(py_version)
        if ver.dev:
            pre_ver = ver.dev
        elif ver.pre:
            pre_ver = ver.pre
        if pre_ver:
            pre_release = [('.', pre_ver[0] if len(pre_ver) > 0 else None,
                            pre_ver[1] if len(pre_ver) > 1 else None)]
        if ver.local:
            local = tuple(itertools.chain.from_iterable(
                (dot, str(_)) for dot, _ in zip('.' * len(ver.local), ver.local)))[1:]
        _LOG.debug('parsing %s %s', type(py_version), py_version)
        return cls(major, minor, patch, pre_release=pre_release, local=local)

    @classmethod
    def from_sem_version(cls, sem_version: t.Union[dict, semver.VersionInfo]):
        """Create version from semantic version object."""
        _LOG.debug('parsing %s %s', type(sem_version), sem_version)
        if isinstance(sem_version, semver.VersionInfo):
            major, minor, patch = sem_version.major, sem_version.minor, sem_version.patch
            pre_release = sem_version.prerelease
            local = sem_version.build
        else:
            major, minor, patch = sem_version['major'], sem_version['minor'], sem_version['patch']
            pre_release = sem_version['prerelease']
            local = sem_version['build']
        if pre_release is not None:
            raise NotImplementedError(sem_version)
        if local is not None:
            local = cls._parse_local_str(f'+{local}')
        return cls(major, minor, patch, pre_release=pre_release, local=local)

    @classmethod
    def from_version(cls, version: 'Version'):
        return cls.from_dict(version.to_dict())

    def __init__(
            self, major: int, minor: t.Optional[int] = None, patch: t.Optional[int] = None, *args,
            pre_release: t.Sequence[
                t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]] = None,
            local: t.Union[str, tuple] = None):
        self._major: t.Optional[int] = None
        self._minor: t.Optional[int] = None
        self._patch: t.Optional[int] = None
        self._pre_release: t.Optional[
            t.List[t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]]] = None
        self._local: t.Optional[t.Sequence[str]] = None

        self.release = major, minor, patch

        if args and pre_release is not None and local is not None:
            raise ValueError(f'args={args}, pre_release={pre_release} and local={local}'
                             f' are all present in {repr(self)}')

        if pre_release is None:
            pre_release, consumed_args = self._get_pre_release_from_args(args)
            if consumed_args > 0:
                args = args[consumed_args:]
            else:
                pre_release = None
        self.pre_release = pre_release

        if args and local is not None:
            raise ValueError(f'args={args} and local={local} are present at the same time'
                             f' in {repr(self)}')

        if local is None:
            if len(args) == 1 and isinstance(args[0], tuple):
                local = args[0]
            elif not args:
                local = None
            else:
                local = args
        elif isinstance(local, str):
            local = (local,)
        self.local = local

    def _get_pre_release_from_args(self, args) -> t.Tuple[
            t.Sequence[t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]], int]:
        pre_release: t.List[t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]] = []
        consumed_args = 0
        if args and isinstance(args[0], tuple):
            for i, arg in enumerate(args):
                if not isinstance(arg, tuple):
                    break
                if len(arg) == 3 and arg[0] in (None, '.', '-'):
                    pre_release.append(arg)
                    consumed_args += 1
                    continue
                if i == len(args) - 1:
                    break
                raise ValueError(f'pre-release segment arg={arg} (index {i} in args={args}'
                                 f' in {repr(self)}) must be a 3-tuple')
        else:
            accumulated: t.List[t.Union[int, str]] = []
            for i, arg in enumerate(args):
                if not accumulated:
                    if arg in (None, '.', '-'):
                        if len(args) < i + 3:
                            raise ValueError(f'expected 3 consecutive values from index {i}'
                                             f' in args={args} in {repr(self)}')
                    else:
                        break
                accumulated.append(arg)
                consumed_args += 1
                if len(accumulated) == 3:
                    pre_release.append(tuple(accumulated))
                    accumulated = []
        return pre_release, consumed_args

    @property
    def release(self) -> t.Tuple[int, t.Optional[int], t.Optional[int]]:
        assert self._major is not None
        return self._major, self._minor, self._patch

    @release.setter
    def release(self, release: t.Tuple[int, t.Optional[int], t.Optional[int]]):
        # major: int, minor: t.Optional[int] = None, patch: t.Optional[int] = None):
        if not isinstance(release, tuple):
            raise TypeError(
                f'release={repr(release)} is of wrong type {type(release)} in {repr(self)}')
        if len(release) != 3:
            raise ValueError(
                f'release={repr(release)} has wrong length {len(release)} in {repr(self)}')

        major, minor, patch = release

        if not isinstance(major, int):
            raise TypeError(f'major={repr(major)} is of wrong type {type(major)} in {repr(self)}')
        if major < 0:
            raise ValueError(f'major={repr(major)} has wrong value in {repr(self)}')
        if minor is not None and not isinstance(minor, int):
            raise TypeError(f'minor={repr(minor)} is of wrong type {type(minor)} in {repr(self)}')
        if minor is not None and minor < 0:
            raise ValueError(f'minor={repr(minor)} has wrong value in {repr(self)}')
        if patch is not None and not isinstance(patch, int):
            raise TypeError(f'patch={repr(patch)} is of wrong type {type(patch)} in {repr(self)}')
        if patch is not None and patch < 0:
            raise ValueError(f'patch={repr(patch)} has wrong value in {repr(self)}')
        if minor is None and patch is not None:
            raise ValueError(f'patch={repr(patch)} is present but not minor in {repr(self)}')

        self._major = major
        self._minor = minor
        self._patch = patch

    @property
    def pre_release(self) -> t.Optional[
            t.Sequence[t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]]]:
        """Pre-release version component."""
        if self._pre_release is None:
            return None
        return self._pre_release.copy()

    @pre_release.setter
    def pre_release(
            self, pre_release: t.Optional[
                t.List[t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]]]):
        if pre_release is None:
            self._pre_release = None
            return

        if not isinstance(pre_release, collections.abc.Sequence):
            raise TypeError(f'pre_release={repr(pre_release)} is of wrong type {type(pre_release)}'
                            f' in {repr(self)}')
        if len(pre_release) == 0:
            raise ValueError(f'pre_release has no elements although it is set in {repr(self)}')

        for pre in pre_release:
            if not isinstance(pre, tuple):
                raise TypeError(
                    f'pre-release part {repr(pre)} is of wrong type {type(pre)} in {repr(self)}')
            if len(pre) != 3:
                raise ValueError(
                    f'pre-release part={repr(pre)} has wrong length {len(pre)} in {repr(self)}')
            pre_separator, pre_type, pre_patch = pre
            self._check_pre_release_parts(pre_separator, pre_type, pre_patch)

        self._pre_release = pre_release

    def _check_pre_release_parts(self, pre_separator, pre_type, pre_patch):
        """Verify that the given pre-release version identifier parts are valid."""
        if pre_separator is not None and not isinstance(pre_separator, str):
            raise TypeError(f'pre_separator={repr(pre_separator)} is of wrong type'
                            f' {type(pre_separator)} in {repr(self)}')
        if pre_separator is not None and pre_separator not in ('-', '.'):
            raise ValueError(f'pre_separator={repr(pre_separator)} has wrong value in {repr(self)}')
        if pre_type is not None and not isinstance(pre_type, str):
            raise TypeError(
                f'pre_type={repr(pre_type)} is of wrong type {type(pre_type)} in {repr(self)}')
        if pre_type is not None and type(self)._pattern_letters.fullmatch(pre_type) is None:
            raise ValueError(f'pre_type={repr(pre_type)} has wrong value in {repr(self)}')
        if pre_patch is not None and not isinstance(pre_patch, int):
            raise TypeError(
                f'pre_patch={repr(pre_patch)} is of wrong type {type(pre_patch)} in {repr(self)}')
        if pre_patch is not None and pre_patch < 0:
            raise ValueError(f'pre_patch={repr(pre_patch)} has wrong value in {repr(self)}')
        if pre_separator is None and pre_type is None and pre_patch is not None:
            raise ValueError(f'neither pre_separator nor pre_type is set'
                             f' but pre_patch={repr(pre_patch)} is in {repr(self)}')
        if pre_separator is not None and pre_type is None and pre_patch is None:
            raise ValueError(f'pre_separator={repr(pre_separator)} is present'
                             f' but neither pre_type nor pre_patch is in {repr(self)}')

    @property
    def has_pre_release(self):
        return self._pre_release is not None

    @property
    def local(self) -> t.Optional[t.Sequence[str]]:
        return self._local

    @local.setter
    def local(self, local: t.Optional[t.Sequence[str]]):
        if local is None:
            self._local = None
            return

        if not isinstance(local, collections.abc.Sequence):
            raise TypeError(f'local={repr(local)} is of wrong type {type(local)} in {repr(self)}')

        if len(local) % 2 != 1:
            raise ValueError(f'local={repr(local)} has wrong length {len(local)} in {repr(self)}')

        for i, part in enumerate(local):
            if not isinstance(part, str):
                raise TypeError(f'local_part or local_separator {repr(part)} is of wrong type'
                                f' {type(part)} in {repr(self)}')
            if i % 2 == 0:
                if type(self)._pattern_alphanumeric.fullmatch(part) is None:
                    raise ValueError(f'local_part={repr(part)} has wrong value in {repr(self)}')
            elif part not in ('-', '.'):
                raise ValueError(f'local_separator={repr(part)} has wrong value in {repr(self)}')

        self._local = tuple(local)

    @property
    def has_local(self):
        return self._local is not None

    def increment(self, component: VersionComponent, amount: int = 1) -> 'Version':
        """Increment a selected version component and return self."""
        if not isinstance(component, VersionComponent):
            raise TypeError(f'component={repr(component)} is of wrong type {type(component)}')
        if not isinstance(amount, int):
            raise TypeError(f'amount={repr(amount)} is of wrong type {type(amount)}')
        if amount < 1:
            raise ValueError(f'amount={amount} has wrong value')

        if component in (VersionComponent.Major, VersionComponent.Minor, VersionComponent.Patch):
            self._increment_release(component, amount)

        elif component is VersionComponent.PrePatch:
            self._increment_pre_path(amount)

        elif component is VersionComponent.DevPatch:
            if self._pre_release is None:
                self._pre_release = []
            if not self._pre_release or self.pre_release_to_tuple(True)[-1][1] != 'dev':
                self._pre_release.append(('.', 'dev', amount))
            else:
                pre_sep, pre_type, pre_patch = self._pre_release[-1]
                if pre_patch is None:
                    pre_patch = amount
                else:
                    pre_patch += amount
                self._pre_release[-1] = (pre_sep, pre_type, pre_patch)
        else:
            raise ValueError(f'incrementing component={repr(component)} is not possible')

        return self

    def _increment_release(self, component: VersionComponent, amount: int):
        if component in (VersionComponent.Major, VersionComponent.Minor):
            if component is VersionComponent.Major:
                assert self._major is not None
                self._major += amount
                if self._minor is not None:
                    self._minor = 0
            else:
                if self._minor is None:
                    self._minor = amount
                else:
                    self._minor += amount

            if self._patch is not None:
                self._patch = 0

        else:
            if self._minor is None:
                self._minor = 0
            if self._patch is None:
                self._patch = amount
            else:
                self._patch += amount

        self._pre_release = None
        self._local = None

    def _increment_pre_path(self, amount):
        if self._pre_release is None or self.pre_release_to_tuple(True)[0][1] != '':
            self._pre_release = [('-', None, amount)]
        else:
            pre_sep, pre_type, pre_patch = self._pre_release[0]
            assert isinstance(pre_patch, int), (type(pre_patch), pre_patch)
            pre_patch += amount
            self.pre_release = [(pre_sep, pre_type, pre_patch)]

    def devel_increment(self, new_commits: int = 1) -> 'Version':
        """Increment version depending on current version number and return self.

        If there's no dev patch, then increment release patch by one and increment dev patch
        by number of new commits.

        If there's a dev patch, then increment it by number of new commits.
        """
        if not self.has_pre_release or self.pre_release_to_tuple(True)[-1][1] != 'dev':
            self.increment(VersionComponent.Patch)
        self.increment(VersionComponent.DevPatch, new_commits)

        return self

    def release_to_str(self) -> str:
        """Get string representation of this version's release component."""
        version_tuple = self._major, self._minor, self._patch
        if _version_tuple_checker(version_tuple, (True, False, False)):
            return '.'.join(str(_) for _ in version_tuple[:1])
        if _version_tuple_checker(version_tuple, (True, True, False)):
            return '.'.join(str(_) for _ in version_tuple[:2])
        if _version_tuple_checker(version_tuple, (True, True, True)):
            return '.'.join(str(_) for _ in version_tuple[:3])
        raise ValueError(f'cannot generate valid version string from {repr(self)}')

    def _pre_release_segment_to_str(self, segment: int) -> str:
        assert self._pre_release is not None
        version_tuple = self._pre_release[segment]
        if _version_tuple_checker(version_tuple, (True, True, False)):
            return '{}{}'.format(*version_tuple[:2])
        if _version_tuple_checker(version_tuple, (True, False, True)):
            return f'{version_tuple[0]}{version_tuple[2]}'
        if _version_tuple_checker(version_tuple, (True, True, True)):
            return '{}{}{}'.format(*version_tuple)
        raise ValueError(f'cannot generate valid version string from {repr(self)}')

    def pre_release_to_str(self) -> str:
        if self._pre_release is None:
            return ''
        return ''.join(self._pre_release_segment_to_str(i)
                       for i, _ in enumerate(self._pre_release))

    def local_to_str(self) -> str:
        if not self._local:
            return ''
        return f'+{"".join(self._local)}'

    def to_str(self) -> str:
        return f'{self.release_to_str()}{self.pre_release_to_str()}{self.local_to_str()}'

    def release_to_tuple(self, sort: bool = False) -> tuple:
        return (0 if sort else None) if self._major is None else self._major, \
            (0 if sort else None) if self._minor is None else self._minor, \
            (0 if sort else None) if self._patch is None else self._patch

    def pre_release_segment_to_tuple(self, segment: int, sort: bool = False) -> tuple:
        assert self._pre_release is not None
        pre_separator, pre_type, pre_patch = self._pre_release[segment]
        return (1 if pre_type is None else 0) if sort else pre_separator, \
            ('' if pre_type is None else pre_type.lower()) if sort else pre_type, \
            (0 if sort else None) if pre_patch is None else pre_patch

    def pre_release_to_tuple(self, sort: bool = False) -> tuple:
        """Create tuple from this version's pre-release component."""
        if self._pre_release is None:
            if sort:
                return ((1, '', 0),)
            return ()
        parts = [self.pre_release_segment_to_tuple(i, sort)
                 for i, _ in enumerate(self._pre_release)]
        return tuple(parts) if sort else tuple(itertools.chain.from_iterable(parts))

    def local_to_tuple(self, sort: bool = False) -> tuple:
        if self._local is None:
            return ()
        return tuple(0 if _ in ('.', '-') else _.lower() for _ in self._local) \
            if sort else self._local

    def to_tuple(self, sort: bool = False) -> tuple:
        return self.release_to_tuple(sort) + self.pre_release_to_tuple(sort) \
            + self.local_to_tuple(sort)

    def to_dict(self) -> dict:
        return {field[1:]: value for field, value in vars(self).items()}

    def to_py_version(self) -> packaging.version.Version:
        return packaging.version.Version(self.to_str())

    def to_sem_version(self) -> semver.VersionInfo:
        return semver.VersionInfo.parse(self.to_str())

    def __repr__(self):
        fields = ', '.join(f'{field[1:]}: {repr(value)}' for field, value in vars(self).items())
        return f'{type(self).__name__}({fields})'

    def __str__(self):
        return self.to_str()

    def __hash__(self):
        return hash(self.to_tuple(sort=True))

    def __eq__(self, other):
        return not self < other and not other < self

    def __ne__(self, other):
        return self < other or other < self

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return not self < other

    def __lt__(self, other):
        if not isinstance(other, Version):
            raise TypeError(f'cannot compare {type(self)} and {type(other)}')

        self_release = self.release_to_tuple(True)
        other_release = other.release_to_tuple(True)
        if self_release != other_release:
            return self_release < other_release

        self_pre_release = self.pre_release_to_tuple(True)
        other_pre_release = other.pre_release_to_tuple(True)
        for self_part, other_part in itertools.zip_longest(
                self_pre_release, other_pre_release, fillvalue=(1, '', 0)):
            if self_part != other_part:
                return self_part < other_part

        self_local = self.local_to_tuple(True)
        other_local = other.local_to_tuple(True)
        for self_part, other_part in zip(self_local, other_local):
            if self_part != other_part:
                return self_part < other_part
        if len(self_local) != len(other_local):
            return len(self_local) < len(other_local)

        return False

    def __le__(self, other):
        return not other < self



"""Tests of version string parsing, generation and comparison."""

import logging
import unittest

import packaging.version
import pkg_resources
import semver

from version_query.version import VersionComponent, Version
from .examples import \
    INIT_CASES, BAD_INIT_CASES, COMPATIBLE_STR_CASES, STR_CASES, case_to_version_tuple, \
    INCREMENT_CASES, DEVEL_INCREMENT_CASES, COMPARISON_CASES_LESS, COMPARISON_CASES_EQUAL

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    maxDiff = None

    def test_from_str(self):
        for version_str, (args, kwargs) in STR_CASES.items():
            version_tuple = case_to_version_tuple(args, kwargs)
            with self.subTest(version_str=version_str, version_tuple=version_tuple):

                py_version = \
                    pkg_resources.parse_version(version_str)  # type: packaging.version.Version
                # self.assertIsInstance(
                #     py_version, packaging.version.Version, msg=(type(py_version), py_version))
                _LOG.debug('packaging parsed version string %s into %s: %s',
                           repr(version_str), type(py_version), py_version)

                try:
                    sem_version = semver.VersionInfo.parse(version_str)  # type: semver.VersionInfo
                except ValueError:
                    _LOG.debug('semver could not parse version string %s', repr(version_str))
                else:
                    _LOG.debug('semver parsed version string %s into %s: %s',
                               repr(version_str), type(sem_version), sem_version)

                self.assertEqual(Version.from_str(version_str).to_tuple(), version_tuple)

    def test_from_str_bad(self):
        with self.assertRaises(ValueError):
            Version.from_str('hello world')

    def test_to_str(self):
        for result, (args, kwargs) in STR_CASES.items():
            with self.subTest(args=args, kwargs=kwargs, result=result):
                self.assertEqual(Version(*args, **kwargs).to_str(), result)

    def test_from_py_version(self):
        for version_str, (args, kwargs) in COMPATIBLE_STR_CASES.items():
            version_tuple = case_to_version_tuple(args, kwargs)
            with self.subTest(version_str=version_str, version_tuple=version_tuple):
                py_version = packaging.version.Version(version_str)
                self.assertEqual(Version.from_py_version(py_version).to_tuple(),
                                 version_tuple, py_version)
                py_version_setuptools = pkg_resources.parse_version(version_str)
                self.assertEqual(Version.from_py_version(py_version_setuptools).to_tuple(),
                                 version_tuple, py_version_setuptools)

    def test_to_py_version(self):
        for version_str, (args, kwargs) in COMPATIBLE_STR_CASES.items():
            version_tuple = case_to_version_tuple(args, kwargs)
            with self.subTest(version_str=version_str, version_tuple=version_tuple):
                version = Version.from_str(version_str)
                py_version = packaging.version.Version(version_str)
                self.assertEqual(version.to_py_version(), py_version,
                                 msg=(version.to_py_version(), py_version))
                # py_version_setuptools = pkg_resources.parse_version(version_str)

    def test_from_sem_version(self):
        for version_str, (args, kwargs) in COMPATIBLE_STR_CASES.items():
            version_tuple = case_to_version_tuple(args, kwargs)
            with self.subTest(version_str=version_str, version_tuple=version_tuple):
                try:
                    sem_version = semver.VersionInfo.parse(version_str)  # type: semver.VersionInfo
                except ValueError:
                    continue
                else:
                    self.assertEqual(Version.from_sem_version(sem_version).to_tuple(),
                                     version_tuple, sem_version)
                sem_version_dict = sem_version.to_dict()
                self.assertEqual(Version.from_sem_version(sem_version_dict).to_tuple(),
                                 version_tuple, sem_version_dict)

    def test_to_sem_version(self):
        for version_str, (args, kwargs) in COMPATIBLE_STR_CASES.items():
            version_tuple = case_to_version_tuple(args, kwargs)
            with self.subTest(version_str=version_str, version_tuple=version_tuple):
                try:
                    sem_version = semver.VersionInfo.parse(version_str)
                except ValueError:
                    continue
                version = Version.from_str(version_str)
                self.assertEqual(version.to_sem_version(), sem_version)

    def test_from_version(self):
        for version_str, args_kwargs_tuple in INIT_CASES.items():
            args, kwargs = args_kwargs_tuple
            with self.subTest(args=args, kwargs=kwargs, version_str=version_str):
                version = Version.from_str(version_str)
                created_version = Version(*args, **kwargs)
                self.assertIsInstance(created_version, Version)
                self.assertEqual(version, created_version)
                version_copy = Version.from_version(created_version)
                self.assertEqual(version, version_copy)
                self.assertEqual(version_copy, created_version)

    def test_init(self):
        for version_str, args_kwargs_tuple in INIT_CASES.items():
            args, kwargs = args_kwargs_tuple
            with self.subTest(args=args, kwargs=kwargs, version_str=version_str):
                version = Version(*args, **kwargs)
                self.assertIsInstance(version, Version)
                self.assertEqual(Version.from_str(version_str), version)
                self.assertIsInstance(version.release, tuple)
                if version.pre_release is not None:
                    self.assertIsInstance(version.pre_release, list)
                if version.has_local:
                    self.assertIsInstance(version.local, tuple)

    def test_init_bad(self):
        for (args, kwargs), exception in BAD_INIT_CASES.items():
            with self.subTest(args=args, kwargs=kwargs, exception=exception):
                with self.assertRaises(exception):
                    Version(*args, **dict(kwargs))
        version = Version(1, 0)
        with self.assertRaises(TypeError):
            version.release = 2
        with self.assertRaises(ValueError):
            version.release = 2, 0
        with self.assertRaises(ValueError):
            version.release = 2, 0, 0, 1
        with self.assertRaises(ValueError):
            version.local = '42', 'and', '43'

    def test_increment(self):
        for (initial_version, args), result_version in INCREMENT_CASES.items():
            with self.subTest(initial_version=initial_version, args=args,
                              result_version=result_version):
                self.assertEqual(Version.from_str(initial_version).increment(*args),
                                 Version.from_str(result_version))

    def test_devel_increment(self):
        for (initial_version, args), result_version in DEVEL_INCREMENT_CASES.items():
            with self.subTest(initial_version=initial_version, args=args,
                              result_version=result_version):
                self.assertEqual(Version.from_str(initial_version).devel_increment(*args),
                                 Version.from_str(result_version))

    def test_increment_bad(self):
        version = Version(1, 0, 0)
        with self.assertRaises(TypeError):
            version.increment(3)
        with self.assertRaises(TypeError):
            version.increment('dev')
        with self.assertRaises(TypeError):
            version.increment(VersionComponent.Minor, '5')
        with self.assertRaises(ValueError):
            version.increment(VersionComponent.Minor, -1)
        with self.assertRaises(ValueError):
            version.increment(VersionComponent.Local)

    def test_compare(self):
        for earlier_version, later_version in COMPARISON_CASES_LESS.items():
            earlier = Version.from_str(earlier_version)
            later = Version.from_str(later_version)
            with self.subTest(earlier_version=earlier_version, later_version=later_version):
                self.assertLess(earlier, later)
                self.assertLessEqual(earlier, later)
                self.assertNotEqual(earlier, later)
                self.assertNotEqual(later, earlier)
                self.assertGreaterEqual(later, earlier)
                self.assertGreater(later, earlier)

        for version, equivalent_version in COMPARISON_CASES_EQUAL.items():
            original = Version.from_str(version)
            equivalent = Version.from_str(equivalent_version)
            with self.subTest(version=version, equivalent_version=equivalent_version):
                self.assertLessEqual(original, equivalent)
                self.assertLessEqual(equivalent, original)
                self.assertEqual(original, equivalent)
                self.assertEqual(equivalent, original)
                self.assertGreaterEqual(original, equivalent)
                self.assertGreaterEqual(equivalent, original)

    def test_compare_bad(self):
        with self.assertRaises(TypeError):
            assert Version(1, 0) < '2.0'

    def test_hash(self):
        for version, equivalent_version in COMPARISON_CASES_EQUAL.items():
            original = Version.from_str(version)
            equivalent = Version.from_str(equivalent_version)
            with self.subTest(version=version, equivalent_version=equivalent_version):
                self.assertEqual(hash(original), hash(equivalent))
                self.assertDictEqual({original: equivalent}, {equivalent: original})



"""Command-line interface of version_query package."""

import argparse
import logging
import os
import pathlib
import sys

from ._version import VERSION
from .version import VersionComponent
from .query import query_folder, predict_folder


def main(args=None, namespace=None) -> None:
    """Run the command-line interface.

    Either query or predict version in a given folder according to the arguments.
    """
    logging_level = getattr(logging, os.environ.get('LOGGING_LEVEL', 'warning').upper())
    logging.basicConfig(level=min(logging_level, logging.WARNING))
    logging.getLogger().setLevel(logging.WARNING)
    logging.getLogger('version_query').setLevel(logging_level)
    parser = argparse.ArgumentParser(
        prog='version_query',
        description='''Tool for querying current versions of Python packages. Use LOGGING_LEVEL
        environment variable to adjust logging level.''',
        epilog='''Copyright 2017-2020 by the contributors, Apache License 2.0,
        https://github.com/mbdevpl/version-query''',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-i', '--increment', action='store_true', help='''output version string for
                        next patch release, i.e. if version is 1.0.3, output 1.0.4''')
    parser.add_argument('-p', '--predict', action='store_true', help='''operate in prediction mode,
                        i.e. assume existence of git repository and infer current version from
                        its tags, history and working tree status''')
    parser.add_argument('path', type=pathlib.Path)
    parser.add_argument('--version', action='version',
                        version=f'{parser.prog} {VERSION},\nPython {sys.version}')
    parsed_args = parser.parse_args(args=args, namespace=namespace)
    if parsed_args.predict and parsed_args.increment:
        raise ValueError(
            'choose one: either increment current version, or predict upcoming version')
    if parsed_args.predict:
        version = predict_folder(parsed_args.path)
    else:
        version = query_folder(parsed_args.path)
    if parsed_args.increment:
        version.increment(VersionComponent.Patch)
    print(version)



"""High-level utility functions for querying, manipulating and generating version information."""

import inspect
import logging
import pathlib

import git

from .version import Version
from .git_query import query_git_repo, predict_git_repo
from .py_query import query_package_folder

_LOG = logging.getLogger(__name__)


def _caller_folder(stack_level: int = 1) -> pathlib.Path:
    """Determine folder in which the caller module of a function is located."""
    frame_info = inspect.getouterframes(inspect.currentframe())[stack_level]
    caller_path = frame_info[1]  # frame_info.filename

    here = pathlib.Path(caller_path).absolute().resolve()
    assert here.is_file(), here
    here = here.parent
    assert here.is_dir(), here
    _LOG.debug('found directory "%s"', here)

    return here


def query_folder(path: pathlib.Path, search_parent_directories: bool = False) -> Version:
    """Determine version of code in a given folder."""
    try:
        return query_git_repo(path, search_parent_directories=search_parent_directories)
    except git.InvalidGitRepositoryError:
        pass
    return query_package_folder(path, search_parent_directories=search_parent_directories)


def query_caller(stack_level: int = 1) -> Version:
    """Determine the version of code associated with the caller of this function."""
    here = _caller_folder(stack_level + 1)
    return query_folder(here, True)


def query_version_str() -> str:
    """Determine the version of code that calls this function and get it as string.

    You can use this utility function like this:

    '''
    from version_query import query_version_str

    VERSION = query_version_str()
    '''
    """
    return query_caller(2).to_str()


def predict_folder(path: pathlib.Path, search_parent_directories: bool = True) -> Version:
    """Predict version of code residing in a given folder."""
    priority_cutoff = 2
    paths = [path] + (list(path.parents)[:priority_cutoff] if search_parent_directories else [])
    for pth in paths:
        try:
            return predict_git_repo(pth, search_parent_directories=False)
        except git.InvalidGitRepositoryError:
            pass
    try:
        return query_package_folder(path, search_parent_directories=False)
    except ValueError:
        pass
    try:
        return predict_git_repo(path, search_parent_directories=search_parent_directories)
    except git.InvalidGitRepositoryError:
        pass
    return query_folder(path, search_parent_directories=search_parent_directories)


def predict_caller(stack_level: int = 1) -> Version:
    """Predict the version of code associated with the caller of this function."""
    here = _caller_folder(stack_level + 1)
    return predict_folder(here, True)


def predict_version_str() -> str:
    """Predict version of the code that calls this function and get it as string.

    You can use this utility function like this:

    '''
    from version_query import predict_version_str

    VERSION = predict_version_str()
    '''
    """
    return predict_caller(2).to_str()



"""Initialization of tests of version_query package."""

import logging
import os

import colorlog

_HANDLER = logging.StreamHandler()
_HANDLER.setFormatter(colorlog.ColoredFormatter(
    '{name} [{log_color}{levelname}{reset}] {message}', style='{'))

logging.basicConfig(level=logging.DEBUG, handlers=[_HANDLER])
logging.getLogger().setLevel(logging.WARNING)
logging.getLogger('version_query').setLevel(
    getattr(logging, os.environ.get('LOGGING_LEVEL', 'debug').upper()))
logging.getLogger('test').setLevel(logging.DEBUG)

if 'EXAMPLE_PROJECTS_PATH' not in os.environ:
    os.environ['EXAMPLE_PROJECTS_PATH'] = '..'



"""Tests of querying tools."""

import contextlib
import importlib
import io
import logging
import os
import pathlib
import sys
import tempfile
import unittest

from version_query.version import VersionComponent, Version
from version_query.git_query import query_git_repo, predict_git_repo
from version_query.py_query import query_metadata_json, query_pkg_info, query_package_folder
from version_query.query import \
    query_folder, query_caller, query_version_str, predict_caller, predict_version_str
from .examples import \
    PY_LIB_DIR, GIT_REPO_EXAMPLES, METADATA_JSON_EXAMPLE_PATHS, PKG_INFO_EXAMPLE_PATHS, \
    PACKAGE_FOLDER_EXAMPLES
from .test_setup import run_module

_LOG = logging.getLogger(__name__)

IGNORED_FOLDER_NAMES = ['opencv']


@contextlib.contextmanager
def temporarily_set_logger_level(logger_name: str, level: int):
    """Change logger level on enter and restore on exit of this context."""
    logger = logging.getLogger(logger_name)
    level_ = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(level_)


def preserve_logger_level(logger_name: str):
    return temporarily_set_logger_level(logger_name, logging.getLogger(logger_name).level)


class Tests(unittest.TestCase):

    def _check_examples_count(self, description, examples):
        lvl = logging.WARNING if len(examples) < 10 else logging.INFO
        _LOG.log(lvl, '%s count: %i', description, len(examples))
        if len(examples) < 5:
            _LOG.warning('%s list: %s', description, examples)
        self.assertGreater(len(examples), 0)

    def test_example_count_checking(self):
        _LOG.warning('%s', PY_LIB_DIR)

        with self.assertRaises(AssertionError):
            self._check_examples_count('test', [])
        self._check_examples_count('test', list(range(1)))
        self._check_examples_count('test', list(range(9)))
        self._check_examples_count('test', list(range(10)))

    def _query_test_case(self, paths, query_function):
        for path in paths:
            if any(_ in path.parts for _ in IGNORED_FOLDER_NAMES):
                continue
            with self.subTest(path=path, query_function=query_function):
                _LOG.debug('testing %s() on %s', query_function.__name__, path)
                try:
                    version = query_function(path)
                except ValueError:
                    _LOG.info('failed to get version from %s', path, exc_info=True)
                else:
                    _LOG.debug('%s: %s', path, version)

    def test_query_git_repo(self):
        self._check_examples_count('git repo', GIT_REPO_EXAMPLES)
        self._query_test_case(GIT_REPO_EXAMPLES, query_git_repo)

    def test_predict_caller_bad(self):
        with tempfile.TemporaryDirectory() as project_path_str:
            with tempfile.NamedTemporaryFile(suffix='.py', dir=project_path_str,
                                             delete=False) as project_file:
                project_file_path = pathlib.Path(project_file.name)
            with project_file_path.open('a', encoding='utf-8') as project_file:
                project_file.write('from version_query.query import predict_caller\n\n\n'
                                   'def caller():\n    predict_caller()\n\n\ncaller()\n')
            sys.path.insert(0, project_path_str)
            _LOG.warning('inserted %s to sys.path', project_path_str)
            _LOG.info('temporary project file path: %s', project_file_path)
            with self.assertRaises(ValueError):
                importlib.import_module(project_file_path.with_suffix('').name)
            sys.path.remove(project_path_str)
            _LOG.warning('removed %s from sys.path', project_path_str)
            project_file_path.unlink()

    def test_predict_git_repo(self):
        self._query_test_case(GIT_REPO_EXAMPLES, predict_git_repo)

    @unittest.skipIf(not METADATA_JSON_EXAMPLE_PATHS, 'no "metadata.json" files found')
    def test_query_metadata_json(self):
        self._check_examples_count('metadata.json', METADATA_JSON_EXAMPLE_PATHS)
        self._query_test_case(METADATA_JSON_EXAMPLE_PATHS, query_metadata_json)

    @unittest.skipIf(not PKG_INFO_EXAMPLE_PATHS, 'no "PKG-INFO" files found')
    def test_query_pkg_info(self):
        self._check_examples_count('PKG-INFO', PKG_INFO_EXAMPLE_PATHS)
        self._query_test_case(PKG_INFO_EXAMPLE_PATHS, query_pkg_info)

    @unittest.skipUnless(
            os.environ.get('TEST_PACKAGING') or os.environ.get('CI'), 'skipping packaging test')
    def test_query_pkg_info_current(self):
        with preserve_logger_level('version_query'):
            run_module('setup', 'build')
        paths = list(pathlib.Path.cwd().glob('*.egg-info/PKG-INFO'))
        self.assertEqual(len(paths), 1)
        path = paths[0]
        version = query_pkg_info(path)
        _LOG.debug('%s: %s', path, version)

    def test_query_pkg_info_bad(self):
        with tempfile.NamedTemporaryFile(delete=False) as bad_file:
            bad_file_path = pathlib.Path(bad_file.name)
        with self.assertRaises(ValueError):
            query_pkg_info(bad_file_path)

        with bad_file_path.open('a', encoding='utf-8') as bad_file:
            bad_file.write('blah blah blah')
        with self.assertRaises(ValueError):
            query_pkg_info(bad_file_path)

        with bad_file_path.open('a', encoding='utf-8') as bad_file:
            bad_file.write('Version: hello world')
        with self.assertRaises(ValueError):
            query_pkg_info(bad_file_path)

        bad_file_path.unlink()

    def test_query_package_folder(self):
        self._check_examples_count('package folder', PACKAGE_FOLDER_EXAMPLES)
        self._query_test_case(PACKAGE_FOLDER_EXAMPLES, query_package_folder)

    @unittest.skipUnless(
            os.environ.get('TEST_PACKAGING') or os.environ.get('CI'), 'skipping packaging test')
    def test_query_package_folder_current(self):
        with preserve_logger_level('version_query'):
            run_module('setup', 'build')
        path = pathlib.Path.cwd().joinpath('version_query')
        version = query_package_folder(path)
        _LOG.debug('%s: %s', path, version)
        self.assertIsInstance(version, Version)

    def test_query_folder(self):
        self._query_test_case(PACKAGE_FOLDER_EXAMPLES, query_folder)

    def test_query_folder_current(self):
        path = pathlib.Path.cwd()
        version = query_folder(path)
        _LOG.debug('%s: %s', path, version)
        self.assertIsInstance(version, Version)

    def test_query_caller(self):
        version = query_caller()
        _LOG.debug('caller: %s', version)
        self.assertIsInstance(version, Version)

    def test_not_as_main(self):  # pylint: disable = no-self-use
        run_module('version_query', run_name=None)

    def test_help(self):
        sio = io.StringIO()
        with contextlib.redirect_stderr(sio):
            with preserve_logger_level('version_query'):
                with self.assertRaises(SystemExit):
                    run_module('version_query')
        _LOG.info('%s', sio.getvalue())

    def test_bad_usage(self):
        sio = io.StringIO()
        with contextlib.redirect_stderr(sio):
            with preserve_logger_level('version_query'):
                with self.assertRaises(ValueError):
                    run_module('version_query', '-p', '-i', '.')
        _LOG.info('%s', sio.getvalue())

    def test_here(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('version_query'):
                run_module('version_query', '.')
        self.assertEqual(sio.getvalue().rstrip(), query_caller().to_str())
        self.assertEqual(sio.getvalue().rstrip(), query_version_str())

    def test_increment_here(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('version_query'):
                run_module('version_query', '-i', '.')
        self.assertEqual(sio.getvalue().rstrip(),
                         query_caller().increment(VersionComponent.Patch).to_str())

    def test_predict_here(self):
        sio = io.StringIO()
        with contextlib.redirect_stdout(sio):
            with preserve_logger_level('version_query'):
                run_module('version_query', '-p', '.')
        self.assertEqual(sio.getvalue().rstrip(), predict_caller().to_str())
        self.assertEqual(sio.getvalue().rstrip(), predict_version_str())



"""Examples for tests."""

import itertools
import os
import pathlib
import platform
import sys
import typing as t

from version_query.version import VersionComponent

_HERE = pathlib.Path(__file__).resolve().parent

_GIT_REPOS_ROOT = pathlib.Path(os.environ['EXAMPLE_PROJECTS_PATH']).absolute()

GIT_REPO_EXAMPLES = list(_ for _ in _GIT_REPOS_ROOT.glob('**/.git') if _.is_dir())


def python_lib_dir() -> pathlib.Path:
    """Get root folder of currently running Python libraries.

    Currently works only for CPython and PyPy.
    """
    lib_dir_parts = [getattr(sys, 'real_prefix', sys.prefix)]
    if platform.python_implementation() == 'CPython':
        lib_dir_parts.append('lib')
        if platform.system() != 'Windows':
            lib_dir_parts.append('python{}.{}'.format(*sys.version_info[:2]))
    else:
        assert platform.python_implementation() == 'PyPy'
        lib_dir_parts += ['lib-python', '{}'.format(*sys.version_info[:1])]

    lib_dir = pathlib.Path(*lib_dir_parts)
    assert lib_dir.is_dir(), lib_dir
    return lib_dir


PY_LIB_DIR = python_lib_dir()

_SYS_DIST_INFOS = list(PY_LIB_DIR.glob('**/*.dist-info'))
_SYS_EGG_INFOS = list(PY_LIB_DIR.glob('**/*.egg-info'))
_USER_DIST_INFOS = [pth for _ in _GIT_REPOS_ROOT.glob('*') for pth in _.glob('*.dist-info')]
_USER_EGG_INFOS = [pth for _ in _GIT_REPOS_ROOT.glob('*') for pth in _.glob('*.egg-info')]
# print(_SYS_DIST_INFOS, _SYS_EGG_INFOS, _USER_DIST_INFOS, _USER_EGG_INFOS)

METADATA_JSON_EXAMPLE_PATHS = list(itertools.chain.from_iterable(
    _.glob('metadata.json') for _ in itertools.chain(_SYS_DIST_INFOS, _USER_DIST_INFOS)))
PKG_INFO_EXAMPLE_PATHS = list(itertools.chain.from_iterable(
    _.glob('PKG-INFO') for _ in itertools.chain(_SYS_EGG_INFOS, _USER_EGG_INFOS)))


def _remove_version_suffixes(path: pathlib.Path) -> t.Optional[pathlib.Path]:
    name = path.with_suffix('').name
    i = name.rfind('-')
    if i > 0:
        name = name[:i]
    path = path.with_name(name)
    return path if path.is_dir() else None


PACKAGE_FOLDER_EXAMPLES = [pth for pth in [
    _remove_version_suffixes(_)
    for _ in itertools.chain(_SYS_DIST_INFOS, _SYS_EGG_INFOS, _USER_DIST_INFOS, _USER_EGG_INFOS)]
                           if pth is not None]
# print(PACKAGE_FOLDER_EXAMPLES)

KWARG_NAMES = ('major', 'minor', 'patch', 'pre_release', 'local')

INIT_CASES: t.Dict[str, t.Tuple[tuple, dict]] = {
    '1': ((1,), {}),
    '1.0': ((1, 0), {}),
    '1.0.0': ((1, 0, 0), {}),
    '1.0.0.rc2': ((1, 0, 0, '.', 'rc', 2), {}),
    '1.0.0.rc3': ((1, 0, 0, ('.', 'rc', 3)), {}),
    '1.0.0.rc2+local': ((1, 0, 0, '.', 'rc', 2, 'local'), {}),
    '1.0.0.rc3+local': ((1, 0, 0, ('.', 'rc', 3), 'local'), {}),
    '1.0.0.rc4+local': ((1, 0, 0, ('.', 'rc', 4), ('local',)), {})}

BAD_INIT_CASES: t.Dict[t.Tuple[tuple, tuple], t.Type[Exception]] = {
    (('spam',), ()): TypeError,
    ((-1,), ()): ValueError,
    ((1, 'ham'), ()): TypeError,
    ((1, -2), ()): ValueError,
    ((1,), (('patch', 13),)): ValueError,
    ((1, 0, 'eggs'), ()): TypeError,
    ((1, 0, -3), ()): ValueError,
    ((1, 0, 0, 0), ()): TypeError,
    ((1, 0, 0, '.', 'dev'), ()): ValueError,
    ((1, 0, 0, '.'), ()): ValueError,
    ((1, 0, 0, '.', 1), ()): ValueError,
    ((1, 0, 0, ('.', None, 1), ('.', 'dev'), ('local',)), ()): ValueError,
    ((1, 0, 0, 0, ), (('pre_release', None), ('local', 'abc'))): ValueError,
    ((1, 0, 0, 0, ), (('pre_release', ((None, 'dev', 0),)), ('local', 'bad'))): ValueError,
    ((5,), (('pre_release', 1),)): TypeError,
    ((5,), (('pre_release', ()),)): ValueError,
    ((5,), (('pre_release', (0,)),)): TypeError,
    ((5,), (('pre_release', ('dev',)),)): TypeError,
    ((5,), (('pre_release', ((None, 'dev'),)),)): ValueError,
    ((6,), (('pre_release', ((42, 'dev', 0),)),)): TypeError,
    ((6,), (('pre_release', (('spam', 'dev', 0),)),)): ValueError,
    ((6,), (('pre_release', ((None, 42, 0),)),)): TypeError,
    ((6,), (('pre_release', ((None, 'd-v', 0),)),)): ValueError,
    ((6,), (('pre_release', ((None, 'dev', 'eggs'),)),)): TypeError,
    ((6,), (('pre_release', ((None, 'dev', -1),)),)): ValueError,
    ((6,), (('pre_release', ((None, None, 1),)),)): ValueError,
    ((6,), (('pre_release', (('.', None, None),)),)): ValueError,
    ((7,), (('local', 0),)): TypeError,
    ((7,), (('local', ('abc', '.')),)): ValueError,
    ((7,), (('local', (0,)),)): TypeError,
    ((7,), (('local', ('a.c',)),)): ValueError,
    ((7,), (('local', ('abc', 1, 'def')),)): TypeError,
    ((7,), (('local', ('abc', 'def')),)): ValueError,
    ((7,), (('local', ('abc', '.', 9)),)): TypeError}

COMPATIBLE_STR_CASES = {
    '0': ((0, None, None), {}),
    '42': ((42, None, None), {}),
    '5+d2a610ba': ((5, None, None), {'local': 'd2a610ba'}),
    '0.1': ((0, 1, None), {}),
    '8.0': ((8, 0, None), {}),
    '16.4': ((16, 4, None), {}),
    '5.0+f753199e': ((5, 0, None), {'local': 'f753199e'}),
    '0.54.0': ((0, 54, 0), {}),
    '1.0.0': ((1, 0, 0), {}),
    '7.0.42+1d7b090a': ((7, 0, 42), {'local': '1d7b090a'}),
    '1.0.0.4': ((1, 0, 0, '.', None, 4), {}),
    '2.0.0.8+cc81cee': ((2, 0, 0, '.', None, 8, 'cc81cee'), {}),
    '4.5.0.dev1234': ((4, 5, 0, '.', 'dev', 1234), {}),
    '1.0.0.rc3': ((1, 0, 0, '.', 'rc', 3), {}),
    '1.0.1.dev0': ((1, 0, 1, '.', 'dev', 0), {}),
    '0.4.4.dev5+84e1d430': ((0, 4, 4, '.', 'dev', 5, '84e1d430'), {}),
    '0.4.4.dev5+20171003.84e1d430': ((0, 4, 4, '.', 'dev', 5, '20171003', '.', '84e1d430'), {})}

INCOMPATIBLE_STR_CASES: t.Dict[str, t.Tuple[tuple, dict]] = {
    '1.0.0-2': ((1, 0, 0, '-', None, 2), {}),
    '1.0.0-0.2': ((1, 0, 0, '-', None, 0, '.', None, 2), {}),
    '4.5.0.dev': ((4, 5, 0, '.', 'dev', None), {})}

STR_CASES = dict(itertools.chain(COMPATIBLE_STR_CASES.items(),
                                 INCOMPATIBLE_STR_CASES.items()))

BAD_STR_CASES = {
    '-1.0.0': ValueError,
    '1.0.0.ekhm_what': ValueError}


def case_to_version_tuple(args, kwargs):
    """Convert parameters of a given test case to a version tuple.

    To be converted are args and kwargs meant for Version.__init__().
    """
    return args + tuple(
        v for _, v in sorted(kwargs.items(), key=lambda _: KWARG_NAMES.index(_[0])))


INCREMENT_CASES = {
    ('1', (VersionComponent.Major,)): '2',
    ('1', (VersionComponent.Minor,)): '1.1',
    ('1', (VersionComponent.Patch, 8)): '1.0.8',
    ('1', (VersionComponent.PrePatch,)): '1-1',
    ('1', (VersionComponent.DevPatch,)): '1.dev1',
    ('1.5', (VersionComponent.Major,)): '2.0',
    ('1.5', (VersionComponent.Minor, 3)): '1.8',
    ('1.5', (VersionComponent.Patch,)): '1.5.1',
    ('1.5', (VersionComponent.PrePatch, 2)): '1.5-2',
    ('1.5', (VersionComponent.DevPatch,)): '1.5.dev1',
    ('1.5.1', (VersionComponent.Major, 3)): '4.0.0',
    ('1.5.1', (VersionComponent.Minor,)): '1.6.0',
    ('1.5.1', (VersionComponent.Patch,)): '1.5.2',
    ('1.5.1', (VersionComponent.PrePatch,)): '1.5.1-1',
    ('1.5.1', (VersionComponent.DevPatch,)): '1.5.1.dev1',
    ('1.5.1-2.4', (VersionComponent.Major,)): '2.0.0',
    ('1.5.1-2.4', (VersionComponent.Minor,)): '1.6',
    ('1.5.1-2.4', (VersionComponent.Patch,)): '1.5.2',
    ('1.5.1-0', (VersionComponent.PrePatch, 5)): '1.5.1-5',
    ('1.5.1-2.4', (VersionComponent.PrePatch,)): '1.5.1-3',
    ('1.5.1-2.4', (VersionComponent.DevPatch,)): '1.5.1-2.4.dev1',
    ('1.5-2.dev', (VersionComponent.PrePatch,)): '1.5-3',
    ('1.5-2.4.dev', (VersionComponent.DevPatch,)): '1.5-2.4.dev1',
    ('1.5-2.dev4', (VersionComponent.PrePatch,)): '1.5-3',
    ('1.5-2.4.dev4', (VersionComponent.DevPatch,)): '1.5-2.4.dev5',
    ('1.5.1-rc', (VersionComponent.PrePatch,)): '1.5.1-1',
    ('1.5.1-rc4', (VersionComponent.PrePatch, 2)): '1.5.1-2',
    ('0.dev', (VersionComponent.DevPatch,)): '0.dev1',
    ('0.dev2', (VersionComponent.DevPatch,)): '0.dev3',
    ('0.3dev', (VersionComponent.DevPatch,)): '0.3dev1',
    ('0.3dev2', (VersionComponent.DevPatch,)): '0.3dev3',
    ('0.3.5dev', (VersionComponent.DevPatch,)): '0.3.5dev1',
    ('0.3.5dev2', (VersionComponent.DevPatch,)): '0.3.5dev3'}

DEVEL_INCREMENT_CASES = {
    ('1', ()): '1.0.1.dev1',
    ('1', (3,)): '1.0.1.dev3',
    ('1.5', ()): '1.5.1.dev1',
    ('1.5', (3,)): '1.5.1.dev3',
    ('1.5.1', ()): '1.5.2.dev1',
    ('1.5.1', (3,)): '1.5.2.dev3',
    ('1.5.1-2.4', ()): '1.5.2.dev1',
    ('1.5.1-2.4', (3,)): '1.5.2.dev3',
    ('0.dev2', ()): '0.dev3',
    ('0.dev2', (3,)): '0.dev5',
    ('0.3.dev2', ()): '0.3.dev3',
    ('0.3.dev2', (3,)): '0.3.dev5',
    ('0.3.5dev2', ()): '0.3.5dev3',
    ('0.3.5dev2', (3,)): '0.3.5dev5',
    ('0.3.5-2.4.dev2', ()): '0.3.5-2.4.dev3',
    ('0.3.5-2.4.dev2', (3,)): '0.3.5-2.4.dev5'}

COMPARISON_CASES_LESS = {
    '0.3-4.4-2.9': '0.3-4.4-2.10',
    '0.3dev': '0.3dev1',
    '0.3rc2': '0.3',
    '0.3': '0.3-2',
    '0.3-2rc5': '0.3-2',
    '0.3-2.dev5': '0.3-2',
    '0.3-2.dev42': '0.3-2',
    '0.3-4': '0.3-4.5',
    '1-1.2.3.4.5.dev4': '1-1.2.3.4.5',
    '1.0.0': '1.0.0+blahblah',
    '1.0.0+aa': '1.0.0+aaa'}

COMPARISON_CASES_EQUAL = {
    '1.0.0': '1.0.0',
    '1': '1.0.0',
    '1.0': '1.0.0.0',
    '1.0.0-0.0.DEV42': '1.0.0.0.0.dev42'}



"""Python package version query tools."""

import logging
import json
import pathlib

from .version import Version

_LOG = logging.getLogger(__name__)


def query_metadata_json(path: pathlib.Path) -> Version:
    """Get version from metadata.json file."""
    with path.open('r') as metadata_file:
        metadata = json.load(metadata_file)
    version_str = metadata['version']
    return Version.from_str(version_str)


def query_pkg_info(path: pathlib.Path) -> Version:
    """Get version from PKG-INFO file."""
    with path.open('r') as pkginfo_file:
        for line in pkginfo_file:
            if line.startswith('Version:'):
                version_str = line.replace('Version:', '').strip()
                return Version.from_str(version_str)
    raise ValueError(path)


def query_package_folder(path: pathlib.Path, search_parent_directories: bool = False) -> Version:
    """Get version from Python package folder."""
    paths = [path]
    if search_parent_directories:
        paths += path.parents
    metadata_json_paths, pkg_info_paths = None, None
    for pth in paths:
        metadata_json_paths = list(pth.parent.glob(f'{pth.name}*.dist-info/metadata.json'))
        pkg_info_paths = list(pth.parent.glob(f'{pth.name}*.egg-info/PKG-INFO'))
        pkg_info_paths += list(pth.parent.glob(f'{pth.name}*.dist-info/METADATA'))
        if len(metadata_json_paths) == 1 and not pkg_info_paths:
            return query_metadata_json(metadata_json_paths[0])
        if not metadata_json_paths and len(pkg_info_paths) == 1:
            return query_pkg_info(pkg_info_paths[0])
    raise ValueError(paths, metadata_json_paths, pkg_info_paths)



"""Tests of adversarial git repos."""

import itertools
import logging
# import unittest

# import git

from version_query.version import VersionComponent, Version
from version_query.git_query import query_git_repo, predict_git_repo
from .test_with_git_repo import GitRepoTests

_LOG = logging.getLogger(__name__)


class Tests(GitRepoTests):
    """Test suite for automated tests of generated git repositories.

    Each case is executed in a fresh empty repository.
    """

    def setUp(self):
        super().setUp()
        self.git_init()

    def test_empty_repo(self):
        with self.assertRaises(ValueError):
            query_git_repo(self.repo_path)
        with self.assertRaises(ValueError):
            predict_git_repo(self.repo_path)

    def test_no_tags(self):
        for i in range(1, 5):
            self.git_commit_new_file()
            with self.assertRaises(ValueError):
                query_git_repo(self.repo_path)
            version = predict_git_repo(self.repo_path)
            self.assertEqual(version.to_str(), f'0.1.0.dev{i}+{self.repo_head_hexsha}')

    def test_only_nonversion_tags(self):
        for i in range(1, 5):
            self.git_commit_new_file()
            self.repo.create_tag(f'commit_{i}')
            with self.assertRaises(ValueError):
                query_git_repo(self.repo_path)
            version = predict_git_repo(self.repo_path)
            self.assertEqual(version.to_str(), f'0.1.0.dev{i}+{self.repo_head_hexsha}')

    def test_inconsistent_tag_prefix(self):
        version = Version.from_str('1.0')
        for _ in range(5):
            for version_tag_prefix in ('v', 'ver', ''):
                self.git_commit_new_file()
                version.increment(VersionComponent.Minor)
                self.repo.create_tag(f'{version_tag_prefix}{version}')
                current_version = query_git_repo(self.repo_path)
                self.assertEqual(version, current_version)
                upcoming_version = predict_git_repo(self.repo_path)
                self.assertEqual(version, upcoming_version)

    def test_nonversion_tags(self):
        version = Version.from_str('0.1.0')
        self.git_commit_new_file()
        self.repo.create_tag(f'v{version}')
        path = self.git_commit_new_file()
        self.git_modify_file(path, commit=True)
        self.repo.create_tag('release1')
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(version, current_version)
        upcoming_version = predict_git_repo(self.repo_path)
        version.increment(VersionComponent.Patch, 1)
        version.increment(VersionComponent.DevPatch, 2)
        version.local = (self.repo_head_hexsha,)
        self.assertEqual(version, upcoming_version)

    def test_too_long_no_tag(self):
        self.git_commit_new_file()
        self.repo.create_tag('v4.0.0')
        path = self.git_commit_new_file()
        for _ in range(1000):
            self.git_modify_file(path, commit=True)
        with self.assertRaises(ValueError):
            query_git_repo(self.repo_path)
        with self.assertRaises(ValueError):
            predict_git_repo(self.repo_path)

    def test_nonversion_merged_branches(self):
        self.git_commit_new_file()
        self.git_commit_new_file()
        self.repo.create_head('devel')
        self.repo.create_head('experimental')
        self.git_commit_new_file()
        self.repo.create_tag('trial')
        self.repo.git.checkout('experimental')
        self.git_commit_new_file()
        self.repo.git.checkout('devel')
        self.git_commit_new_file()
        self.repo.create_tag('error')
        self.repo.git.checkout('master')
        self.repo.git.merge('devel')
        self.repo.git.merge('experimental')
        self.git_commit_new_file()
        with self.assertRaises(ValueError):
            query_git_repo(self.repo_path)
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'0.1.0.dev6+{self.repo_head_hexsha}')

    def test_invalid_version_tags(self):
        for i in range(1, 3):
            self.git_commit_new_file()
            self.repo.create_tag(f'v{i}.0.0')
            self.git_commit_new_file()
            self.repo.create_tag(f'version_{i + 1}.0.0')

            current_version = query_git_repo(self.repo_path)
            self.assertEqual(current_version.to_str(), f'{i}.0.0')
            upcoming_version = predict_git_repo(self.repo_path)
            current_version.devel_increment(1)
            current_version.local = (self.repo_head_hexsha,)
            self.assertEqual(current_version, upcoming_version)

    def test_dirty_repo(self):
        version = Version.from_str('0.9.0')
        for new_commit, add in itertools.product((False, True), (False, True)):
            path = self.git_commit_new_file()
            self.repo.create_tag(f'v{version}')
            if new_commit:
                self.git_commit_new_file()
            self.git_modify_file(path, add=add)
            current_version = query_git_repo(self.repo_path)
            self.assertEqual(version, current_version)
            upcoming_version = predict_git_repo(self.repo_path)
            self.assertLess(version, upcoming_version)
            if new_commit:
                current_version.devel_increment(1)
                current_version.local = (self.repo_head_hexsha,)
                self.assertTupleEqual(current_version.local, upcoming_version.local[:1])
                local_prefix = f'+{self.repo_head_hexsha}.dirty'
            else:
                local_prefix = '+dirty'
            self.assertTrue(upcoming_version.local_to_str().startswith(local_prefix),
                            msg=upcoming_version.local_to_str())
            self.assertLess(current_version, upcoming_version)
            version.increment(VersionComponent.Patch)
            self.assertLess(upcoming_version, version)

    def test_nonlatest_commit(self):
        self.git_commit_new_file()
        self.repo.create_tag('v0.1.0')
        self.git_commit_new_file()
        self.repo.create_head('devel')
        self.repo.git.checkout('devel')
        self.git_commit_new_file()
        self.repo.create_tag('v0.2.0')
        self.git_commit_new_file()
        self.repo.git.checkout('master')
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(current_version.to_str(), '0.1.0')
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'0.1.1.dev1+{self.repo_head_hexsha}')
        self.repo.git.checkout('devel')
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(current_version.to_str(), '0.2.0')
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'0.2.1.dev1+{self.repo_head_hexsha}')

    def test_tags_on_merged_branches(self):
        self.git_commit_new_file()
        self.repo.create_head('devel')
        self.git_commit_new_file()
        self.repo.create_tag('v0.2.0')
        self.git_commit_new_file()
        self.git_commit_new_file()
        self.git_commit_new_file()
        self.repo.git.checkout('devel')
        self.git_commit_new_file()
        self.repo.create_tag('v0.1.0')
        self.git_commit_new_file()
        self.repo.git.checkout('master')
        self.repo.git.merge('devel')
        self.git_commit_new_file()
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(current_version.to_str(), '0.2.0')
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'0.2.1.dev5+{self.repo_head_hexsha}')

    def test_tag_on_merged_branch(self):
        self.git_commit_new_file()
        self.repo.create_head('devel')
        self.git_commit_new_file()
        self.repo.git.checkout('devel')
        self.git_commit_new_file()
        self.repo.create_tag('v1.0.0')
        self.repo.git.checkout('master')
        self.repo.git.merge('devel')
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(current_version.to_str(), '1.0.0')
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'1.0.1.dev1+{self.repo_head_hexsha}')

    def test_many_versions_on_one_commit(self):
        self.git_commit_new_file()
        self.repo.create_tag('v0.1.0')
        self.repo.create_tag('v0.3.0')
        self.repo.create_tag('v0.2.0')
        self.git_commit_new_file()
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(current_version.to_str(), '0.3.0')
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'0.3.1.dev1+{self.repo_head_hexsha}')

    def test_version_decreased(self):
        self.git_commit_new_file()
        self.repo.create_tag('v0.2.0')
        self.git_commit_new_file()
        self.repo.create_tag('v0.1.0')
        self.git_commit_new_file()
        current_version = query_git_repo(self.repo_path)
        self.assertEqual(current_version.to_str(), '0.1.0')
        upcoming_version = predict_git_repo(self.repo_path)
        self.assertEqual(upcoming_version.to_str(), f'0.1.1.dev1+{self.repo_head_hexsha}')



"""Setup script for version_query package."""

import setup_boilerplate


class Package(setup_boilerplate.Package):
    """Package metadata."""

    name = 'version-query'
    description = 'Zero-overhead package versioning for Python.'
    url = 'https://github.com/mbdevpl/version-query'
    author = 'Mateusz Bysiek, John Vandenberg'
    maintainer = 'Mateusz Bysiek'
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Version Control',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: System :: Software Distribution',
        'Topic :: Utilities',
        'Typing :: Typed']
    keywords = [
        'automation', 'continous integration', 'git', 'releasing', 'semantic versioning', 'tagging',
        'versioning']


if __name__ == '__main__':
    Package.setup()



"""Tests for setup scripts."""

import importlib
import itertools
import os
import pathlib
import runpy
import subprocess
import sys
import tempfile
import types
import typing as t
import unittest

__version__ = '2022.08.27'


def run_program(*args, glob: bool = False) -> None:
    """Run subprocess with given args. Use path globbing for each arg that contains an asterisk."""
    if glob:
        cwd = pathlib.Path.cwd()
        args = tuple(itertools.chain.from_iterable(
            list(str(_.relative_to(cwd)) for _ in cwd.glob(arg)) if '*' in arg else [arg]
            for arg in args))
    try:
        subprocess.run(args, check=True)
    except subprocess.CalledProcessError as err:
        raise AssertionError(f'execution of {args} failed') from err


def run_pip(*args, **kwargs) -> None:
    run_program(sys.executable, '-m', 'pip', *args, **kwargs)


def run_module(name: str, *args, run_name: str = '__main__') -> None:
    backup_sys_argv = sys.argv
    sys.argv = [name + '.py'] + list(args)
    runpy.run_module(name, run_name=run_name)
    sys.argv = backup_sys_argv


def import_module(name: str = 'setup') -> types.ModuleType:
    setup_module = importlib.import_module(name)
    return setup_module


def import_module_member(module_name: str, member_name: str) -> t.Any:
    module = import_module(module_name)
    return getattr(module, member_name)


CLASSIFIERS_LICENSES = (
    'License :: OSI Approved :: Python License (CNRI Python License)',
    'License :: OSI Approved :: Python Software Foundation License',
    'License :: Other/Proprietary License',
    'License :: Public Domain')

CLASSIFIERS_PYTHON_VERSIONS = tuple("""Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.0
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3 :: Only""".splitlines())

CLASSIFIERS_PYTHON_IMPLEMENTATIONS = tuple("""Programming Language :: Python :: Implementation
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: Jython
Programming Language :: Python :: Implementation :: PyPy
Programming Language :: Python :: Implementation :: Stackless""".splitlines())

CLASSIFIERS_VARIOUS = (
    'Framework :: IPython',
    'Topic :: Scientific/Engineering',
    'Topic :: Sociology',
    'Topic :: Security :: Cryptography',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Version Control :: Git',
    'Topic :: System',
    'Topic :: Utilities')

CLASSIFIERS_LICENSES_TUPLES = tuple((_,) for _ in CLASSIFIERS_LICENSES) + ((),)

CLASSIFIERS_PYTHON_VERSIONS_COMBINATIONS = tuple((_,) for _ in CLASSIFIERS_PYTHON_VERSIONS)

CLASSIFIERS_PYTHON_IMPLEMENTATIONS_TUPLES = tuple((_,) for _ in CLASSIFIERS_PYTHON_IMPLEMENTATIONS)

CLASSIFIERS_VARIOUS_COMBINATIONS = tuple(itertools.combinations(
    CLASSIFIERS_VARIOUS, len(CLASSIFIERS_VARIOUS) - 1)) + (CLASSIFIERS_VARIOUS,)

ALL_CLASSIFIERS_VARIANTS = [
    licenses + versions + implementations + various
    for licenses in CLASSIFIERS_LICENSES_TUPLES
    for versions in CLASSIFIERS_PYTHON_VERSIONS_COMBINATIONS
    for implementations in CLASSIFIERS_PYTHON_IMPLEMENTATIONS_TUPLES
    for various in CLASSIFIERS_VARIOUS_COMBINATIONS]

LINK_EXAMPLES = [
    (None, 'setup.py', True), ('this file', 'setup.py', True), (None, 'test/test_setup.py', True),
    (None, 'test/test_setup.py#L98', True), ('line 5 of this file', 'setup.py#L5', True),
    (None, 'http://site.com', False), (None, '../something/else', False), (None, 'no.thing', False),
    (None, '/my/abs/path', False)]


def get_package_folder_name():
    """Attempt to guess the built package name."""
    cwd = pathlib.Path.cwd()
    directories = [
        path for path in cwd.iterdir() if pathlib.Path(cwd, path).is_dir()
        and pathlib.Path(cwd, path, '__init__.py').is_file() and path.name != 'test']
    assert len(directories) == 1, directories
    return directories[0].name


class UnitTests(unittest.TestCase):
    """Test basic functionalities of the setup boilerplate."""

    def test_find_version(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        result = find_version(get_package_folder_name())
        self.assertIsInstance(result, str)

    def test_find_packages(self):
        find_packages = import_module_member('setup_boilerplate', 'find_packages')
        results = find_packages()
        self.assertIsInstance(results, list)
        for result in results:
            self.assertIsInstance(result, str)

    def test_requirements(self):
        parse_requirements = import_module_member('setup_boilerplate', 'parse_requirements')
        results = parse_requirements()
        self.assertIsInstance(results, list)
        self.assertTrue(all(isinstance(result, str) for result in results), msg=results)

    def test_requirements_empty(self):
        parse_requirements = import_module_member('setup_boilerplate', 'parse_requirements')
        with tempfile.NamedTemporaryFile('w', delete=False) as reqs_file:
            pass
        results = parse_requirements(reqs_file.name)
        self.assertIsInstance(results, list)
        self.assertEqual(len(results), 0)
        os.remove(reqs_file.name)

    def test_requirements_comments(self):
        parse_requirements = import_module_member('setup_boilerplate', 'parse_requirements')
        reqs = ['# comment', 'numpy', '', '# another comment', 'scipy', '', '# one more comment']
        with tempfile.NamedTemporaryFile('w', delete=False) as reqs_file:
            for req in reqs:
                print(req, file=reqs_file)
        results = parse_requirements(reqs_file.name)
        self.assertIsInstance(results, list)
        self.assertGreater(len(results), 0)
        self.assertLess(len(results), len(reqs))
        os.remove(reqs_file.name)

    def test_python_versions(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        for variant in ALL_CLASSIFIERS_VARIANTS:
            with self.subTest(variant=variant):
                result = find_required_python_version(variant)
                if result is not None:
                    self.assertIsInstance(result, str)

    def test_python_versions_combined(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifiers = [
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.5']
        req = find_required_python_version(classifiers)
        self.assertEqual(req, '>=3.5')

    def test_python_versions_reversed(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifiers = [
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6']
        req = find_required_python_version(classifiers)
        self.assertEqual(req, '>=3.4')
        req = find_required_python_version(reversed(classifiers))
        self.assertEqual(req, '>=3.4')

    def test_python_versions_none(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        result = find_required_python_version([])
        self.assertIsNone(result)

    def test_python_versions_many_only(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifiers = [
            'Programming Language :: Python :: 2 :: Only',
            'Programming Language :: Python :: 3 :: Only']
        with self.assertRaises(ValueError):
            find_required_python_version(classifiers)

    def test_python_versions_conflict(self):
        find_required_python_version = import_module_member(
            'setup_boilerplate', 'find_required_python_version')
        classifier_variants = [
            ['Programming Language :: Python :: 2.7',
             'Programming Language :: Python :: 3 :: Only'],
            ['Programming Language :: Python :: 2 :: Only',
             'Programming Language :: Python :: 3.0']]
        for classifiers in classifier_variants:
            with self.assertRaises(ValueError):
                find_required_python_version(classifiers)


class PackageTests(unittest.TestCase):
    """Test methods of Package class."""

    def test_try_fields(self):
        package = import_module_member('setup_boilerplate', 'Package')

        class Package(package):  # pylint: disable=too-few-public-methods
            name = 'package name'
            description = 'package description'
        self.assertEqual(Package.try_fields('name', 'description'), 'package name')
        self.assertEqual(Package.try_fields('bad_field', 'description'), 'package description')
        with self.assertRaises(AttributeError):
            self.assertIsNone(Package.try_fields())
        with self.assertRaises(AttributeError):
            Package.try_fields('bad_field', 'another_bad_field')

    def test_parse_readme(self):
        package = import_module_member('setup_boilerplate', 'Package')

        class Package(package):  # pylint: disable=too-few-public-methods
            name = 'package name'
            description = 'package description'
            version = '1.2.3.4'
            url = 'https://github.com/example'

        with tempfile.NamedTemporaryFile('w', suffix='.md', delete=False) as temp_file:
            temp_file.write('test test test')
        result, content_type = Package.parse_readme(temp_file.name)
        os.remove(temp_file.name)
        self.assertIsInstance(result, str)
        self.assertIsInstance(content_type, str)

        prefix = 'https://github.com/example/blob/v1.2.3.4/'
        for name, link, done in LINK_EXAMPLES:
            name = '' if name is None else name + ' '
            text = f'Please see `{name}<{link}>`_ for details.'
            with tempfile.NamedTemporaryFile('w', suffix='.rst', delete=False) as temp_file:
                temp_file.write(text)
            result, content_type = Package.parse_readme(temp_file.name)
            os.remove(temp_file.name)
            self.assertIsInstance(result, str)
            self.assertIsInstance(content_type, str)
            if not done:
                self.assertEqual(result, text)
                continue
            if name == '':
                name = link + ' '
            self.assertIn(f'`{name}<{prefix}{link}>`_', result)

    def test_prepare(self):
        package = import_module_member('setup_boilerplate', 'Package')

        version_ = '1.2.3.4.5.6.7'
        long_description_ = 'long package description'

        class Package(package):  # pylint: disable=too-few-public-methods, missing-docstring
            name = 'package name'
            version = version_
            description = 'package description'
            long_description = long_description_
            packages = []
            install_requires = []
            python_requires = ''

        self.assertEqual(Package.version, version_)
        self.assertEqual(Package.long_description, long_description_)
        Package.prepare()
        self.assertEqual(Package.version, version_)
        self.assertEqual(Package.long_description, long_description_)

        del Package.long_description
        del Package.packages
        del Package.install_requires
        del Package.python_requires
        Package.prepare()

        del Package.version
        with self.assertRaises(FileNotFoundError):
            Package.prepare()


@unittest.skipUnless(os.environ.get('TEST_PACKAGING') or os.environ.get('CI'),
                     'skipping packaging tests for actual package')
class IntergrationTests(unittest.TestCase):
    """Test if the boilerplate can actually create a valid package."""

    pkg_name = get_package_folder_name()

    def test_build_binary(self):
        run_module('setup', 'bdist')
        self.assertTrue(os.path.isdir('dist'))

    def test_build_wheel(self):
        run_module('setup', 'bdist_wheel')
        self.assertTrue(os.path.isdir('dist'))

    def test_build_source(self):
        run_module('setup', 'sdist', '--formats=gztar,zip')
        self.assertTrue(os.path.isdir('dist'))

    def test_install_code(self):
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip('install', '--prefix', temporary_folder, '.')
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_install_source_tar(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        version = find_version(self.pkg_name)
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip(
                'install', '--prefix', temporary_folder, f'dist/*-{version}.tar.gz', glob=True)
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_install_source_zip(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        version = find_version(self.pkg_name)
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip('install', '--prefix', temporary_folder, f'dist/*-{version}.zip', glob=True)
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_install_wheel(self):
        find_version = import_module_member('setup_boilerplate', 'find_version')
        version = find_version(self.pkg_name)
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip(
                'install', '--prefix', temporary_folder, f'dist/*-{version}-*.whl', glob=True)
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_pip_error(self):
        with self.assertRaises(AssertionError):
            run_pip('wrong_pip_command')

    def test_setup_do_nothing_or_error(self):
        run_module('setup', 'wrong_setup_command', run_name='__not_main__')
        with self.assertRaises(SystemExit):
            run_module('setup', 'wrong_setup_command')
